from __future__ import annotations

from typing import Generic, TypeVar

_T = TypeVar("_T")


class DoubleLinkedListNode(Generic[_T]):

    __slots__ = (
        "_value",
        "next",
        "prev",
    )

    def __init__(self, value: _T) -> None:
        self._value = value
        self.next: DoubleLinkedListNode[_T] | None = None
        self.prev: DoubleLinkedListNode[_T] | None = None

    @property
    def value(self) -> _T:
        return self._value

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DoubleLinkedListNode):
            return False
        if self is other:
            return True
        return self._value == other._value  # type: ignore[no-any-return]

    def __repr__(self) -> str:  # pragma: no cover
        return f"{self.__class__.__name__}(value={self._value})"



from __future__ import annotations

from datetime import timedelta

import pytest

from ttlru_map import TTLMap, TTLMapInvalidConfigError


@pytest.mark.parametrize(
    ("ttl", "max_size", "update_ttl_on_get"),
    [
        (timedelta(seconds=1), 1, True),
        (timedelta(seconds=1), 1, False),
        (timedelta(seconds=1), None, True),
        (None, 1, False),
    ],
)
def test_init__ok(ttl: timedelta | None, max_size: int | None, update_ttl_on_get: bool):
    TTLMap(ttl=ttl, max_size=max_size, update_ttl_on_get=update_ttl_on_get)


@pytest.mark.parametrize(
    ("ttl", "max_size", "update_ttl_on_get"),
    [
        (None, None, False),
        (None, None, True),
        (timedelta(seconds=0), 1, True),
        (timedelta(seconds=-1), 1, True),
        (timedelta(seconds=1), 0, False),
        (timedelta(seconds=1), -1, True),
        (None, 1, True),
    ],
)
def test_init_error(ttl: timedelta | None, max_size: int | None, update_ttl_on_get: bool):
    with pytest.raises(TTLMapInvalidConfigError):
        TTLMap(ttl=ttl, max_size=max_size, update_ttl_on_get=update_ttl_on_get)



from ttlru_map._linked_list import DoubleLinkedListNode


def test_ll_value():
    value = 1
    node: DoubleLinkedListNode[int] = DoubleLinkedListNode(value=value)
    assert node.value == value


def test_ll_equal():
    value = 1
    node1: DoubleLinkedListNode[int] = DoubleLinkedListNode(value=value)
    node2: DoubleLinkedListNode[int] = DoubleLinkedListNode(value=value)
    assert node1 == node2


def test_ll_not_equal():
    value = 1
    node1: DoubleLinkedListNode[int] = DoubleLinkedListNode(value=value)
    node2: DoubleLinkedListNode[int] = DoubleLinkedListNode(value=value + 1)
    assert node1 != node2


def test_ll_not_equal__different_type():
    value = 1
    node: DoubleLinkedListNode[int] = DoubleLinkedListNode(value=value)
    assert node != value



from datetime import timedelta

from ttlru_map import TTLMap


def test_iter():
    d = TTLMap(ttl=timedelta(seconds=1000))
    assert list(d) == []

    d["key1"] = "value"
    assert list(d) == ["key1"]

    d["key2"] = "value"
    assert list(d) == ["key1", "key2"]



from datetime import timedelta
from unittest.mock import patch

from tests.utils import LockMock
from ttlru_map import TTLMap
from ttlru_map._linked_list import DoubleLinkedListNode
from ttlru_map._ttl_map import _DictValue, _LinkedListValue


def test_set__first():
    d = TTLMap(ttl=timedelta(seconds=1000))
    lock_mock = LockMock()
    d._lock = lock_mock
    key = 1
    value = 2
    time_ = 10
    expected_node = DoubleLinkedListNode(value=_LinkedListValue(time_=time_, key=key))
    with patch("time.time", return_value=time_) as time_mock, patch.object(
        TTLMap,
        "_setitem",
        wraps=d._setitem,
    ) as setitem_mock, patch.object(
        TTLMap,
        "_update_by_ttl",
        wraps=d._update_by_ttl,
    ) as update_by_ttl_mock, patch.object(
        TTLMap,
        "_update_by_size",
        wraps=d._update_by_size,
    ) as update_by_size_mock:
        d[key] = value
        time_mock.assert_called_once()
        setitem_mock.assert_called_once_with(key, value, time_)
        update_by_ttl_mock.assert_called_once_with(current_time=time_)
        update_by_size_mock.assert_called_once()
        lock_mock.__enter__.assert_called_once()
        lock_mock.__exit__.assert_called_once()
        assert d[key] == value
        assert d._dict == {key: _DictValue(value=value, node=expected_node)}
        assert d._ll_head == expected_node
        assert d._ll_end == expected_node


def test_set__second():
    d = TTLMap(ttl=timedelta(seconds=1000))
    head_key = 1
    head_value = 2
    d[head_key] = head_value
    head_item = d._dict[head_key]
    lock_mock = LockMock()
    d._lock = lock_mock
    new_key = 5
    new_value = 20
    time_ = 10
    expected_node = DoubleLinkedListNode(value=_LinkedListValue(time_=time_, key=new_key))
    with patch("time.time", return_value=time_) as time_mock, patch.object(
        TTLMap,
        "_setitem",
        wraps=d._setitem,
    ) as setitem_mock, patch.object(
        TTLMap,
        "_update_by_ttl",
        wraps=d._update_by_ttl,
    ) as update_by_ttl_mock, patch.object(
        TTLMap,
        "_update_by_size",
        wraps=d._update_by_size,
    ) as update_by_size_mock:
        d[new_key] = new_value
        time_mock.assert_called_once()
        setitem_mock.assert_called_once_with(new_key, new_value, time_)
        update_by_ttl_mock.assert_called_once_with(current_time=time_)
        update_by_size_mock.assert_called_once()
        lock_mock.__enter__.assert_called_once()
        lock_mock.__exit__.assert_called_once()
        assert d[new_key] == new_value
        end_node = d._dict[new_key].node
        assert expected_node == end_node
        assert d._dict == {
            head_key: head_item,
            new_key: _DictValue(value=new_value, node=end_node),
        }
        assert d._ll_head is head_item.node
        assert d._ll_end is end_node
        assert head_item.node.next == end_node
        assert end_node.prev == head_item.node



from datetime import timedelta

from ttlru_map import TTLMap


def test_delitem():
    d = TTLMap(ttl=timedelta(seconds=1000))
    key = 1
    value = 1
    d[key] = value

    d._delitem(d._dict[key])
    assert d._dict == {}
    assert d._ll_head is None
    assert d._ll_end is None



from datetime import timedelta
from unittest.mock import patch

import pytest

from tests.utils import LockMock
from ttlru_map import TTLMap
from ttlru_map._linked_list import DoubleLinkedListNode
from ttlru_map._ttl_map import _DictValue, _LinkedListValue


def test_delete():
    d = TTLMap(ttl=timedelta(seconds=1000))
    lock_mock = LockMock()
    d._lock = lock_mock
    key = 1
    value = 2
    time_ = 10

    node = DoubleLinkedListNode(value=_LinkedListValue(time_=time_, key=key))
    item = _DictValue(node=node, value=value)
    d._dict[key] = item
    d._ll_head = node
    d._ll_end = node

    with patch.object(TTLMap, "_delitem", wraps=d._delitem) as delitem_mock, patch.object(
        TTLMap,
        "_update_by_ttl",
        wraps=d._update_by_ttl,
    ) as update_by_ttl_mock:
        del d[key]
        assert d._dict == {}
        assert d._ll_head is None
        assert d._ll_end is None

        delitem_mock.assert_called_once_with(item)
        update_by_ttl_mock.assert_called_once_with()


def test_delete__item_not_found():
    d = TTLMap(ttl=timedelta(seconds=1000))

    with pytest.raises(KeyError):
        del d[1]


def test_delete__item_expired():
    ttl = timedelta(seconds=100)
    d = TTLMap(ttl=ttl)
    key = 1
    value = 2
    time_ = 10

    node = DoubleLinkedListNode(value=_LinkedListValue(time_=time_, key=key))
    d._dict[key] = _DictValue(node=node, value=value)
    d._ll_head = node
    d._ll_end = node

    del d[key]
    assert d._dict == {}
    assert d._ll_head is None
    assert d._ll_end is None



from datetime import timedelta
from unittest.mock import patch

import pytest

from tests.utils import LockMock
from ttlru_map import TTLMap
from ttlru_map._ttl_map import DoubleLinkedListNode, _DictValue, _LinkedListValue


@pytest.mark.parametrize("update_ttl_on_get", [True, False])
def test_get(update_ttl_on_get: bool):
    d = TTLMap(ttl=timedelta(seconds=1000), update_ttl_on_get=update_ttl_on_get)
    lock_mock = LockMock()
    d._lock = lock_mock
    key = 1
    value = 2
    time_ = 10

    node = DoubleLinkedListNode(value=_LinkedListValue(time_=time_, key=key))
    d._dict[key] = _DictValue(node=node, value=value)
    d._ll_head = node
    d._ll_end = node

    with patch("time.time", return_value=time_) as time_mock, patch.object(
        TTLMap,
        "_setitem",
        wraps=d._setitem,
    ) as setitem_mock, patch.object(TTLMap, "_update_by_ttl", wraps=d._update_by_ttl) as update_by_ttl_mock:
        assert d[key] == value
        time_mock.assert_called_once()
        update_by_ttl_mock.assert_called_once_with(current_time=time_)
        if update_ttl_on_get:
            setitem_mock.assert_called_once_with(key, value, time_)
        else:
            setitem_mock.assert_not_called()


def test_get__item_not_found():
    d = TTLMap(ttl=timedelta(seconds=1000))

    with pytest.raises(KeyError):
        _ = d[1]


def test_get__item_expired():
    ttl = timedelta(seconds=100)
    d = TTLMap(ttl=ttl)
    key = 1
    value = 2
    time_ = 10

    node = DoubleLinkedListNode(value=_LinkedListValue(time_=time_, key=key))
    d._dict[key] = _DictValue(node=node, value=value)
    d._ll_head = node
    d._ll_end = node

    with patch("time.time", return_value=time_ + ttl.total_seconds() + 1):  # noqa: SIM117
        with pytest.raises(KeyError):
            _ = d[key]
